#include "tree.h"
#define MAXNAME 11
#define MAXNUMLEN 40

/* Structure for tree (in tree.h)
struct tree {
  int leavesnum; * number of leaves *
  int branchnum; * number of branches, including trivial ones *
  char **leaf; * leaves' names *
  int **branch; * branches as strings of 0 and 1 *
  unsigned int rooted; * 1 - rooted, 0 - unrooted *
  unsigned int root; * index of the root branch *
  unsigned int phylogram; * 1 if lengths are really lengths, 0 otherwise *
  float *length; * branch lengths *
  float rootlocation; * distance from the root to the node from the side of leaf #1 (index 0) *
}

* Structure for distance matrix (in tree.h) *
struct distancematrix {
  int speciesnum; * number of species *
  char **species; * species' names *
  float **distance; * distances *
}
*/

struct tree readbrackets(char *brackets) /* converting string with Newick to rooted tree */
{
  struct tree result;
  char c;
  char tmpstr[2];
  char lenstr[MAXNUMLEN];
  unsigned int i, j, k, s;
  unsigned int flag, nname;
  int stacklen, maxstacklen = 1;
  unsigned int leafi, branchi, stacki, rooti, rootj; 
  int **newbranchstack;

  result.leavesnum = 0;
  result.branchnum = 0;
  result.leaf = (char**)malloc(sizeof(char*));
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*));
  result.length = (float*)malloc(sizeof(float));
  newbranchstack = (int**)malloc(sizeof(int*));

  result.rooted = 0;
  result.phylogram = 0;

  i = 0;
  c = brackets[i];
  flag = 0;
  stacklen = 0;
  while ( (c != ';') && (c != '\0') ) {

    if ( c == ',' ) {
      if ( flag == 2 && strlen(lenstr) > 0 ) { /* branch length is ready */
        result.length[branchi] = atof(lenstr);
      }
      flag = 0; /* wait for a new leaf or new branch */
    }

    else if ( c == '(' ) { /* opening new branch */
      if (flag != 0) {
        perror("Wrong Newick format!");
        exit(1);
      }
      stacklen++;
      /* printf("stacklen = %d, %d leaves, %d branches\n", stacklen, result.leavesnum, result.branchnum); */
      if (stacklen > maxstacklen) {
        maxstacklen = stacklen;
        newbranchstack = (int**)realloc(newbranchstack, sizeof(int*)*maxstacklen);
      }
      stacki = stacklen - 1;
      newbranchstack[stacki] = (int*)malloc(sizeof(int)*result.leavesnum);
      for (k = 0; k < result.leavesnum; k++) {
        newbranchstack[stacki][k] = 0;
      }
    } /* if c == '(' */

    else if ( c == ')' ) { /* new branch is ready */
      if ( flag == 2 && strlen(lenstr) > 0 ) { /* branch length is ready */
        result.length[branchi] = atof(lenstr);
      }
      flag = 3;
      stacklen--;
      if ( stacklen < 0 ) {
        perror("Unbalanced brackets");
        exit(1);
      }
      result.branchnum++;
      branchi = result.branchnum - 1;
      result.branch = (unsigned int**) realloc(result.branch, sizeof(unsigned int*)*result.branchnum);
      result.branch[branchi] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
      for (k = 0; k < result.leavesnum; k++) {
        result.branch[branchi][k] = newbranchstack[stacklen][k];
      } /* for */
      free(newbranchstack[stacklen]);
      result.length = (float*) realloc(result.length, sizeof(float)*result.branchnum);
      result.length[branchi] = 1.0;
    }  /* if c == ')' */

    else if ( c == ':' ) {
      if ( flag == 0 || flag == 2 ) {
        perror("Wrong Newick format!");
        exit(1);
      }
      flag = 2; /* wait for a branch length */
      strcpy(lenstr, "");
    }

    else if ( flag == 0 && !isspace(c) ) { /* regard c as the first symbol of a new leaf name */
      result.leavesnum++;     /* new leaf */
      result.branchnum++;     /* new trivial (one-leaf) branch */
      leafi = result.leavesnum - 1; /* index of the current leaf */
      branchi = result.branchnum - 1; /* index of the current branch */
      result.leaf = (char **) realloc(result.leaf, sizeof(char *)*result.leavesnum);
      result.leaf[leafi] = (char *) malloc(sizeof(char)*MAXNAME);  /* memory for the name of the current leaf */
      result.leaf[leafi][0] = c;
      result.leaf[leafi][1] = '\0';
      
      result.branch = (unsigned int **) realloc(result.branch, sizeof(unsigned int *)*result.branchnum);
      result.length = (float*) realloc(result.length, sizeof(float)*result.branchnum);
      result.branch[branchi] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
      for (j = 0; j < branchi; j++) {
        result.branch[j] = (unsigned int *) realloc(result.branch[j], sizeof(unsigned int)*result.leavesnum);
        result.branch[j][leafi] = 0; /* increasing leaf array for all branches */
      }
      for ( k = 0; k < leafi; k++) {
        result.branch[branchi][k] = 0; 
      }
      result.branch[branchi][leafi] = 1; /* new branch discriminates the new leaf from all old leaves */
      result.length = (float *) realloc (result.length, sizeof(float)*result.branchnum);
      result.length[branchi] = 1.0; /* length is not read yet */

      for (j = 0; (int)j < stacklen; j++) {
        newbranchstack[j] = (int*)realloc(newbranchstack[j], sizeof(int)*result.leavesnum);
        newbranchstack[j][leafi] = 1;
      }

      flag = 1; /* in a leaf name */
      nname = 1; /* name length */
    }  /* if c is the first symbol of new leaf name */

    else if ( flag == 1 && !isspace(c) ) { /* continuation of leaf name */
      nname++;
      if (nname < MAXNAME) {
        tmpstr[0] = c;
        tmpstr[1] = '\0';
        strcat (result.leaf[leafi], tmpstr);
      } /* if */
      else {
        printf("Warning: the name %s... is too long!\n", result.leaf[leafi]); 
      }
    } /* if c is a part of the name */

    else if ( flag == 2 ) { /* branch length */
      if ( c == '.' || isdigit(c) ) {
        tmpstr[0] = c;
        tmpstr[1] = '\0';
        if (strlen(lenstr) + 2 < MAXNUMLEN) {
          strcat(lenstr,tmpstr);
        }
        else {
          perror("Too long number!");
          exit(1);
        }
      } /* if c is a part of a number (is a digit or the decimal point) */
    } /* if flag == 2 */

    i++;
    c = brackets[i];
  } /* while */
  if ( c == '\0' ) {
    perror("No \';\' at the end of string!");
    exit(1);
  }
  if ( stacklen > 0 ) {
    perror("Error: unbalanced brackets");
    exit(1);
  }

  /* removing equal branches and the trivial branch */
  s = 0;
  i = result.branchnum - 1;
  for (k = 0; k < result.leavesnum; k++) {
    s += result.branch[i][k];
  }
  if (s == result.leavesnum) {
    result.branchnum--;
  }
  else {
    perror("No outer bracket pair!");
    exit(1);
  }
  rooti = 0;
  rootj = 0;
  for (i = 1; i < result.branchnum && rooti == 0; i++) {
    for (j = 0; j < i && rooti == 0; j++) {
      s = 0;
      for (k = 0; k < result.leavesnum; k++) {
        if ( result.branch[i][k] == result.branch[j][k] ) {
          s++;
        }
      }
      if (s == 0) { /* branches rooti and rootj are equal <=> coincide with the root */
        rooti = i;
        rootj = j;
      }
    }
  }
  if (rooti != 0) { /* the root was found */
    result.rooted = 1;
    result.root = rooti;
    result.length[rooti] += result.length[rootj];
    result.rootlocation = result.length[rootj];
    result.branchnum--;
    for (j = rootj; j < result.branchnum; j++) { /* elimination of the branch rootj */
      result.branch[j] = result.branch[j+1];
      result.length[j] = result.length[j+1];
    }
    result.root--; /* because rooti > rootj*/
  }

  return result;
} /* readbrackets */


char *writebrackets(struct tree intree, int writelengths) {  /* converting tree to Newick */
  char *result;
  unsigned int s = 0, i, j, k, kk, n, p, u, branchi, leafi, branchnumber, minbranchvol, flag, rbkk, root;
  unsigned int nextbranchi;
  unsigned int **branches;
  float *lengths;
  int *branchindex;
  unsigned int *leftbrackets;
  unsigned int *rightbrackets;
  float **poslengths;
  unsigned int *branchvol;
  unsigned int *branchflag;
  char tmpstr[9];

  
  for (k = 0; k < intree.leavesnum; k++) {
    s += strlen(intree.leaf[k]); /* number of symbols in leaf's name */
  }
  s += 2*(intree.branchnum + 1); 
      /* for each inner branch (incl. two root branches): a pair of brackets;
         for each outer branch: comma and space */
  if (writelengths) {
    s += 10*(intree.branchnum + 1); /* length of branch in the format :0.0000 (up to 999.9999) */
  }
  s += 4; /* outer brackets, semicolumn and the end of string */

  result = (char*)malloc(sizeof(char)*s);
  if ( result == NULL ) {
    perror("writebrackets: malloc");
    exit(1);
  }

  branchindex = (int*)malloc(sizeof(int)*intree.leavesnum);
  leftbrackets = (unsigned int*)malloc(sizeof(unsigned int)*intree.leavesnum);
  rightbrackets = (unsigned int*)malloc(sizeof(unsigned int)*intree.leavesnum);
  poslengths = (float**)malloc(sizeof(float*)*intree.leavesnum);

  for ( k = 0; k < intree.leavesnum; k++ ) {
    branchindex[k] = -1;
    leftbrackets[k] = rightbrackets[k] = 0;
    poslengths[k] = (float*) malloc(sizeof(float));
  }   

  if ( intree.rooted ) {
    root = intree.root;
    branchnumber = intree.branchnum + 1;
  }
  else {
    root = 0;
    branchnumber = intree.branchnum;
  }

  branches = (unsigned int**) malloc(sizeof(unsigned int*)*branchnumber);
  lengths = (float*) malloc(sizeof(float)*branchnumber);
  for ( i = 0; i < branchnumber; i++ ) {
    branches[i] = (unsigned int*)malloc(sizeof(unsigned int)*intree.leavesnum);
  }
  for ( j = 0; j < intree.leavesnum; j++) {
    branches[root][j] = intree.branch[root][j];
  }
  for ( j = 0; j < intree.leavesnum; j++) {
    if ( intree.rooted ) {
      branches[branchnumber-1][j] = 1 - intree.branch[root][j]; /* the additional branch is the root branch in reversed noation */
    }
  }
  if ( intree.rooted ) {
    if ( intree.branch[root][0] == 1 ) {
      lengths[root] = intree.rootlocation;
      lengths[branchnumber-1] = intree.length[root] - intree.rootlocation;
    }
    else {
      lengths[root] = intree.length[root] - intree.rootlocation;
      lengths[branchnumber-1] = intree.rootlocation;
    }
  }
  else {
    lengths[root] = intree.length[root];
  }

  for ( i = 0; i < intree.branchnum; i++) {
    if ( i != root ) {
      flag = 0;
      j = 0;
      while ( (j < intree.leavesnum) && (flag == 0) ) { /* seeking for a leaf j such that i is between j and root */
        if ( branchbetween(intree, i, j, root) ) {
          flag = 1;
        }
        else {
          j++;
        }
      } /* end while ( now i is between j and root ) */
      if ( intree.branch[i][j] == 1 ) {
        for ( k = 0; k < intree.leavesnum; k++ ) {
          branches[i][k] = intree.branch[i][k];
        }
      } /* if i has 1 in j */
      else { /* reverse i */
        for ( k = 0; k < intree.leavesnum; k++ ) {
          branches[i][k] = 1 - intree.branch[i][k];
        }
      }
      lengths[i] = intree.length[i];
    } /* if i != root */
  } /* for all branches i */

  if ( !intree.rooted ) {
    flag = 0;
    for ( k = 0; (k < intree.leavesnum) && (flag == 0); k++ ) {
      n = 0;
      for ( i = 0; i < branchnumber; i++ ) {
        n += branches[i][k];
      }
      if ( n == 0) flag = 1;
    }
    if ( flag == 1 ) {
      for ( k = 0; k < intree.leavesnum; k++ ) {
        branches[root][k] = 1 - branches[root][k];
      }
    }
  }

  branchvol = (unsigned int*)malloc(sizeof(unsigned int)*branchnumber);
  branchflag = (unsigned int*)malloc(sizeof(unsigned int)*branchnumber);

  for ( i = 0; i < branchnumber; i++) {
    branchvol[i] = 0;
    branchflag[i] = 0;
    for (k = 0; k < intree.leavesnum; k++) {
      branchvol[i] += branches[i][k]; 
    }
  }
  
  n = 0;
  while (n < branchnumber) { /* while not all branches are processed */
    i = 0;
    while ( (branchflag[i] == 1) && (i < branchnumber) )i++;
    if ( i == branchnumber) {
      perror("Error 9");
      exit(9);
    }
    nextbranchi = i;
    while ( branchflag[nextbranchi] == 0 ) { /* while the closest outer branch to i is unproceeded */
      i = nextbranchi;
      minbranchvol = intree.leavesnum; /* definitely greater than any branch volume */
      for ( j = 0; j < branchnumber; j++) {
        if ( j != i && branchvol[i] < branchvol[j] && branchvol[j] < minbranchvol ) {
          flag = 0;
          for (k = 0; k < intree.leavesnum && flag == 0; k++) { /* check if i is a subset of j */
            if ( branches[i][k] > branches[j][k] ) flag = 1;
          }
          if ( flag == 0) { /* i is a subset of j */
            nextbranchi = j;
            minbranchvol = branchvol[j];
          }
        }
      }
      if ( nextbranchi == i ) { /* no outer branches at all */
        branchflag[i] = 1; /* processing i */
        n++;
        k = 0;
        while ( branchindex[k] != -1 ) k++;
        for ( kk = 0; kk < branchvol[i]; kk++) {
          branchindex[k+kk] = i;
        }
        if ( branchvol[i] > 1 ) {
          leftbrackets[k]++;
          kk = k + branchvol[i] - 1;
          rightbrackets[kk]++;
          rbkk = rightbrackets[kk] + 1;
          poslengths[kk] = (float*) realloc(poslengths[kk], sizeof(float)*rbkk);
          for ( p = 1; p < rbkk-1; p++ ) {
            poslengths[kk][rbkk-p] = poslengths[kk][rbkk-p-1];
          }
          poslengths[kk][1] = lengths[i];
        }
        else {
          poslengths[k][0] = lengths[i];
        }
      }
    } /* end while */

    if ( branchflag[i] == 0 ) { /* the closest outer branch to i exists and is yet processed */
      branchflag[i] = 1;
      n++;
      for ( k = 0; (branchindex[k] != (int)nextbranchi) && (k < intree.leavesnum); k++);
      for ( kk = 0; kk < branchvol[i]; kk++) {
        branchindex[k+kk] = i;
      }
      if ( branchvol[i] > 1 ) {
        leftbrackets[k]++;
        kk = k + branchvol[i] - 1;
        rightbrackets[kk]++;
        rbkk = rightbrackets[kk] + 1;
        poslengths[kk] = (float*) realloc(poslengths[kk], sizeof(float)*rbkk);
        for ( p = 1; p < rbkk-1; p++ ) {
          poslengths[kk][rbkk-p] = poslengths[kk][rbkk-p-1];
        }
        poslengths[kk][1] = lengths[i];
      }
      else {
        poslengths[k][0] = lengths[i];
      }
    } /* if */
  } /* while n < branchnumber */

  strcpy(result, "(");
  for ( k = 0; k < intree.leavesnum; k++ ) {
    branchi = branchindex[k];
    
    for ( u = 0; u < leftbrackets[k]; u++) {
      strcat(result, "(");
    }

    for ( leafi = 0; branches[branchi][leafi] == 0 && leafi < intree.leavesnum; leafi++);
    strcat(result, intree.leaf[leafi]);
    if (writelengths) {
      sprintf(tmpstr,":%6.4f", poslengths[k][0]);
      strcat(result, tmpstr);
    }
    if ( strlen(result) > s ) {
      printf("Result is too long: %s, s = %d\n", result, s);
      exit(7);
    }

    if (writelengths) {
      for ( u = 0; u < rightbrackets[k]; u++) {
        strcat(result, ")");
        sprintf(tmpstr,":%6.4f", poslengths[k][u+1]);
        strcat(result, tmpstr);
        if ( strlen(result) > s ) {
          printf("Result is too long: %s, s = %d\n", result, s);
          exit(7);
        }
      }
    }
    else {
      for ( u = 0; u < rightbrackets[k]; u++) {
        strcat(result, ")");
      }
    }
    
    if ( k < intree.leavesnum -1 ) { 
      strcat(result, ", "); /* comma after a node (except of the last one */
    }
  }
  strcat(result, ");");
  
  return result;
} /* writebrackets */

struct tree midpoint(struct tree intree) {  /* rooting to midpoint */
  struct tree result;
  unsigned int i, j, k, leaf1, leaf2, pathlen;
  unsigned int *path;
  float dist, maxdist, sum;

  if ( intree.phylogram ) {
    /* find the most distant pair of leaves */
    maxdist = 0.0;
    leaf1 = 0;
    leaf2 = 0;
    for ( j = 1; j < intree.leavesnum; j++) {
      for ( k = 0; k < j; k++) {
        dist = treedistance(intree, j, k);
        if (dist > maxdist) {
          maxdist = dist;
          leaf1 = k;
          leaf2 = j;
        }  
      }
    }
    /* printf("midpoint: maxdist = %6.4f; the most distant leaves are %s and %s\n", maxdist, intree.leaf[leaf1], intree.leaf[leaf2]); */
    if ( maxdist == 0 ) {
      fprintf(stderr, "Midpoint: all distances are zeroes, midpoint cannot be computed.\n");
      result = treecopy(intree);
      result.root = 0;
      result.rootlocation = 0.0;
    }
    else {
      result.leavesnum = intree.leavesnum;
      result.branchnum = intree.branchnum;
      result.rooted = 1;
      result.phylogram = 1;
      result.leaf = (char**)malloc(sizeof(char*)*intree.leavesnum);
      for ( i = 0; i < intree.leavesnum; i++) {
        result.leaf[i] = (char*)malloc(sizeof(char)*(strlen(intree.leaf[i]) + 1));
        strcpy(result.leaf[i], intree.leaf[i]);
      }
      result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*intree.branchnum);
      result.length = (float*)malloc(sizeof(float)*intree.branchnum);
      for ( j = 0; j < intree.branchnum; j++) {
        result.branch[j] = (unsigned int*) malloc(sizeof(unsigned int)*intree.leavesnum);
        result.length[j] = intree.length[j];
      }
      path = treepath(intree, leaf1, leaf2);
      pathlen = combdistance(intree, leaf1, leaf2);
      sum = 0;
      for ( j = 0; j < pathlen && 2*sum < maxdist; j++ ){
        sum += intree.length[path[j]];
      }
      result.root = path[j-1];
      /* printf("midpoint: the root is at the branch no.%d (%d-th in the path)\n", path[j-1] + 1, j); */
      /* making branches */
      for ( i = 0; i < intree.branchnum; i++ ) {
        if ( branchbetween(intree, i, 0, j ) ) { /* the branch is between the leaf number 0 and the root; thus the sign of 0 leaf in it should be 1 */ 
          if ( intree.branch[i][0] == 1 ) {
            for ( k = 0; k < intree.leavesnum; k++ ) {
              result.branch[i][k] = intree.branch[i][k];
            }
          }
          else {
            for ( k = 0; k < intree.leavesnum; k++ ) {
              result.branch[i][k] = 1 - intree.branch[i][k];
            }
          }
        }
        else {
          if ( intree.branch[i][0] == 0 ) {
            for ( k = 0; k < intree.leavesnum; k++ ) {
              result.branch[i][k] = intree.branch[i][k];
            }
          }
          else {
            for ( k = 0; k < intree.leavesnum; k++ ) {
              result.branch[i][k] = 1 - intree.branch[i][k];
            }
          }
        }
      }
      if (result.branch[result.root][leaf2] == result.branch[result.root][0] ) {
        result.rootlocation = sum - maxdist/2; 
      }
      else {
        result.rootlocation = result.length[result.root] - (sum - maxdist/2);
      }
    } /* if maxdist > 0 */
  }
  else {
    fprintf(stderr, "Midpoint: the input tree is not a phylogram, midpoint does not make sence!\n");
    result = treecopy(intree);
  }

  return result;
} /* midpoint */


float treedistance(struct tree intree, unsigned int leaf1, unsigned int leaf2) { /* find the distance among the tree between two leafs */
  float result;
  unsigned int j;

  if ( leaf1 < intree.leavesnum && leaf2 < intree.leavesnum ) {
    if ( !intree.phylogram ) {
      fprintf(stderr, "Treedistance warning: input tree is not a phylogram!\n");
    }

    result = 0.0;
    for ( j = 0; j < intree.branchnum; j++ ) {
      if ( intree.branch[j][leaf1] != intree.branch[j][leaf2] ) {
        result += intree.length[j];
      }
    }
  }
  else { /* at least one of indices is out of range */ 
    result = 0.0;
    fprintf(stderr, "Treedistance: input index is out of range!\n");
  }

  return result;
} /* treedistance */

unsigned int combdistance(struct tree intree, unsigned int leaf1, unsigned int leaf2) { /* find the distance among the tree between two leafs */
  int result;
  unsigned int j;

  if ( leaf1 < intree.leavesnum && leaf2 < intree.leavesnum ) {
    result = 0;
    for ( j = 0; j < intree.branchnum; j++ ) {
      if ( intree.branch[j][leaf1] != intree.branch[j][leaf2] ) {
        result++;
      }
    }
  }
  else { /* at least one of indices is out of range */
    result = 0;
    fprintf(stderr, "Treedistance: input index is out of range!\n");
  }

  return result;
} /* combdistance */

unsigned int *treepath(struct tree intree, int unsigned leaf1, int unsigned leaf2) { /* list of branches forming path from leaf1 to leaf2 */
  unsigned int *result;
  unsigned int cd;
  unsigned int i, j, p, pathnum;

  cd = combdistance(intree, leaf1, leaf2);
  result = (unsigned int*)malloc(sizeof(unsigned int)*cd);

  pathnum = 0;
  for ( i = 0; i < intree.branchnum; i++ ) {
    if ( intree.branch[i][leaf1] != intree.branch[i][leaf2] ) {
      if ( pathnum == 0 ) {
        pathnum = 1;
        result[0] = i;
      }
      else {
        for ( j = 0; j < pathnum && branchbetween(intree, result[j], leaf1, i); j++);
        pathnum++;
        for ( p = 1; p < pathnum - j; p++) {
          result[pathnum-p] = result[pathnum-p-1];
        }
        result[j] = i;
      }
    }
  }

  return result;
} /* treepath */

int branchbetween(struct tree intree, unsigned int branch, unsigned int leaf, unsigned int otherbranch) { 
/* if branch is on the path from leaf to otherbranch, then 1, else 0 */
  int result;
  unsigned int j;

  if ( branch < intree.branchnum && leaf < intree.leavesnum && otherbranch < intree.branchnum ) {
    result = 1;
    for ( j = 0; j < intree.leavesnum && result; j++ ) {
      if ( (intree.branch[branch][j] == intree.branch[branch][leaf]) 
        && (intree.branch[otherbranch][j] != intree.branch[otherbranch][leaf]) ) result = 0;
    }
  }
  else {
    fprintf(stderr, "Branchbetween: indices out of range\n");
    result = 0;
  }
  return result;
}  /* branchbetween */

struct tree treecopy(struct tree intree) { /* copy tree */
  struct tree result;
  unsigned int i, j;

  result.leavesnum = intree.leavesnum;
  result.branchnum = intree.branchnum;
  result.rooted = intree.rooted;
  result.root = intree.root;
  result.phylogram = intree.phylogram;
  result.rootlocation = intree.rootlocation;
  result.leaf = (char**)malloc(sizeof(char*)*intree.leavesnum);
  for ( i = 0; i < intree.leavesnum; i++) {
    result.leaf[i] = (char*)malloc(sizeof(char)*(strlen(intree.leaf[i]) + 1));
    strcpy(result.leaf[i], intree.leaf[i]);
  }
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*intree.branchnum);
  result.length = (float*)malloc(sizeof(float)*intree.branchnum);
  for ( j = 0; j < intree.branchnum; j++) {
    result.branch[j] = (unsigned int*) malloc(sizeof(unsigned int)*intree.leavesnum);
    for ( i = 0; i < intree.leavesnum; i++) {
      result.branch[j][i] = intree.branch[j][i];
    }
    result.length[j] = intree.length[j];
  }
  return result;
} /* treecopy */

struct tree addleaf(struct tree intree, unsigned int branchindex, char *newleaf) { /*  adding new leaf to the branch with the given index */
  struct tree result;
  unsigned int i, j;

  if ( branchindex >= intree.branchnum ) {
    result.leavesnum = 0;
  }
  else {
    result = treecopy(intree);

    result.leavesnum++;
    result.branchnum += 2;
    result.leaf = (char**)realloc(result.leaf, sizeof(char*)*result.leavesnum);
    result.leaf[result.leavesnum-1] = (char*) malloc(sizeof(char)*(strlen(newleaf)+1));
    strcpy(result.leaf[result.leavesnum-1], newleaf);
    result.branch = (unsigned int**)realloc(result.branch, sizeof(unsigned int*)*result.branchnum);
    result.length = (float*)realloc(result.length, sizeof(float)*result.branchnum);

    for ( i = 0; i < intree.branchnum; i++) {
      result.branch[i] = (unsigned int*)realloc(result.branch[i], sizeof(unsigned int)*result.leavesnum);

      if ( i != branchindex ) {
        for ( j = 0; j < intree.leavesnum && !branchbetween(intree, i, j, branchindex); j++);
        result.branch[i][result.leavesnum-1] = 1 - result.branch[i][j];
      }
    } /* for */

    result.branch[branchindex][intree.leavesnum] = 1;
    result.branch[intree.branchnum] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
    for ( j = 0; j < intree.leavesnum; j++) {
      result.branch[intree.branchnum][j] = result.branch[branchindex][j];
    }
    result.branch[intree.branchnum][intree.leavesnum] = 0;
    result.length[branchindex] = intree.length[branchindex]/2;
    result.length[intree.branchnum] = intree.length[branchindex]/2;

    result.branch[intree.branchnum+1] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
    for ( j = 0; j < intree.leavesnum; j++) {
      result.branch[intree.branchnum+1][j] = 0;
    }
    result.branch[intree.branchnum+1][intree.leavesnum] = 1;
    result.length[intree.branchnum+1] = 1.0;
  }
  result.rooted = 0;

  return result;
}

struct tree nni(struct tree intree, unsigned int branchindex, unsigned int which) {
  struct tree result;
  unsigned int i, j, k;
  unsigned int volume, v00, v01, v10, v11, first, second, flag, flag0, flag1;
  unsigned int four[4];

  if ( branchindex >= intree.branchnum ) {
    result.leavesnum = 0;
    result.branchnum = 0;
  }
  else {
    volume = 0;
    for ( k = 0; k < intree.leavesnum; k++ ) {
      volume += intree.branch[branchindex][k];
    }
    if ( volume < 2 || volume > intree.leavesnum - 2 ) {
      result.leavesnum = 0;
      result.branchnum = 0;
    }
    else {
      result = treecopy(intree);

       /* rearrangement of all branches to the outside of branchindex */
      for ( i = 0; i < intree.branchnum; i++ ) if ( i != branchindex ){
        v00 = v01 = v10 = v11 = 1;
        for ( k = 0; k < intree.leavesnum; k++ ) {
          if ( intree.branch[i][k] == 0 && intree.branch[branchindex][k] == 0 ) v00 = 0;
          else if ( intree.branch[i][k] == 0 && intree.branch[branchindex][k] == 1 ) v01 = 0;
          else if ( intree.branch[i][k] == 1 && intree.branch[branchindex][k] == 0 ) v10 = 0; 
          else if ( intree.branch[i][k] == 1 && intree.branch[branchindex][k] == 1 ) v11 = 0;
        }
        if ( v00 + v01 + v10 + v11 > 1 || v00 + v01 + v10 + v11 == 0) {
          fprintf(stderr, "nni for branch %d, something goes wrong with the branch %d\n", branchindex+1, i+1);
          fprintf(stderr, "v00 = %d, v01 = %d, v10 = %d, v11 = %d\n", v00, v01, v10, v11);
          exit(1);
        }
        if ( v00 || v01 ) {
          for ( k = 0; k < intree.leavesnum; k++ ) {
            result.branch[i][k] = 1 - intree.branch[i][k];
          }
        }
      }
      
      /* determination of two pairs of adjacent branches */
      first = second = 0;
      for ( i = 0; i < intree.branchnum - 1 && second == 0; i++ ) if ( i != branchindex ) {
        for ( j = i+1; j < intree.branchnum && second == 0; j++ ) if ( j != branchindex ) {
          flag = 0;
          flag0 = 0;
          flag1 = 0;
          for ( k = 0; k < intree.leavesnum && flag == 0; k++ ) {
            if ( result.branch[i][k] + result.branch[j][k] == 2 ) flag = 1;
            else if ( result.branch[branchindex][k] == 0 ) {
              if ( result.branch[i][k] + result.branch[j][k] > 0 ) {
                flag1 = 1;
                flag = flag0;
              }
              else {
                flag0 = 1;
                flag = flag1;
              }
            }
            else  { /* intree.branch[branchindex][k] == 1 */
              if ( result.branch[i][k] + result.branch[j][k] > 0 ) {
                flag0 = 1;
                flag = flag1;
              }
              else {
                flag1 = 1;
                flag = flag0;
              }
            }       
          }
          if ( flag == 0 ) {
            if ( first == 0) {
              four[0] = i;
              four[1] = j;
              first = 1;
            }
            else {
              four[2] = i;
              four[3] = j;
              second = 1;
            }
          }
        } /* for j */
      } /* for i */

      if ( which == 1 ) {
        for ( k = 0; k < intree.leavesnum; k++ ) {
          result.branch[branchindex][k] = result.branch[four[0]][k] + result.branch[four[2]][k];
        }
      }
      else {
        for ( k = 0; k < intree.leavesnum; k++ ) {
          result.branch[branchindex][k] = result.branch[four[0]][k] + result.branch[four[3]][k];
        }
      }
    } /* else (the branch _branchindex_ is an inner one) */
  } /* else (the branch is not out of range) */
  return result; 
}


float globallength(struct tree intree){  /* sum of lengths of all branches  */
  float result;
  int i;

  result = 0.0;
  for ( i = 0; i < intree.branchnum; i++ ) {
    result += intree.length[i];
  }

  return result;
}



struct distancematrix readphylipmatrix(FILE *infile) { /* read phylip distance matrix */
  struct distancematrix result;
  char line[74];
  unsigned int i, j, k;

  fgets(line, 72, infile);
  if ( !isdigit(line[4]) ) {
    result.speciesnum = 0; /* Wrong format */
    return result;
  }
  sscanf(line,"%d", &(result.speciesnum));
  if ( result.speciesnum < 2 ) {
    result.speciesnum = 0;
  }
  else {
    result.species = (char**) malloc ( sizeof(char*)*result.speciesnum );
    result.distance = (float**) malloc(sizeof(float*)*result.speciesnum);
    for ( i = 0; i < result.speciesnum; i++ ) {
      result.species[i] = (char*)malloc( sizeof(char)*11 );
      result.distance[i] = (float*) malloc(sizeof(float)*result.speciesnum);
      fgets(line, 73, infile);
      if ( line[strlen(line) - 1] != '\n') {
        result.speciesnum = 0;
        return result;
      }
      if ( line[strlen(line)-2] == '\r' ) {
        line[strlen(line)-2] = '\n';
        line[strlen(line)-1] = '\0';
      }
      if ( result.speciesnum < 7 && strlen(line) - 1 != 10*(result.speciesnum + 1)) {
        result.speciesnum = 0; /* Wrong format */
        return result;
      } 
      sscanf(line, "%s", result.species[i]);
      j = 0;
      while ( j < result.speciesnum && j < 6 ) {
        sscanf(line + 10*(j+1), "%f", &(result.distance[i][j]) );
        j++;
      }
      if ( j < result.speciesnum ) { 
        fgets(line, 73, infile);
        if ( line[strlen(line) - 1] != '\n') {
          result.speciesnum = 0;
          return result;
        }
        if ( line[strlen(line)-2] == '\r' ) {
          line[strlen(line)-2] = '\n';
          line[strlen(line)-1] = '\0';
        }
        if ( result.speciesnum < 14 && strlen(line) - 1 != 10*(result.speciesnum - 6)) {
          result.speciesnum = 0; /* Wrong format */
          return result;
        }
        k = 0;
        j = 0;
        while ( 6 + 7*k + j < result.speciesnum ) {
          if ( j == 7 ) {
            fgets(line, 73, infile);
            if ( line[strlen(line)-1] != '\n') {
              result.speciesnum = 0;
              return result;
            }
            if ( line[strlen(line)-2] == '\r' ) {
              line[strlen(line)-2] = '\n';
              line[strlen(line)-1] = '\0';
            }
            if ( result.speciesnum < 7*(k+3) && strlen(line) - 1 != 10*(result.speciesnum - 13 - 7*k)) {
              result.speciesnum = 0; /* Wrong format */
              return result;
            }
            j = 0;
            k++;
          }
          sscanf(line + 10*j, "%f", &(result.distance[i][6+7*k+j]) );
          j++;
        }
      }
    }
  }
  return result;
} /* readphylipmatrix */

struct distancematrix readembossmatrix(FILE *infile) {
  struct distancematrix result;
  char *line;
  char tmpline[8];
  char addline[513];
  unsigned int maxlength = 512;
  unsigned int i, j, k, p, lenline, controlnumber;

  line = (char*)malloc(sizeof(char)*(maxlength + 1));
  fgets(line, 512, infile);
  if ( strncmp(line, "Distance Matrix", 15) != 0 ) {
    result.speciesnum = 0;
  }
  else {
    while( line[0] != '\t' && !feof(infile) ) {
      fgets(line, 512, infile);
    }
    if ( feof(infile) ) {
      result.speciesnum = 0;
    }
    else {
      while ( line[strlen(line)-1] != '\n' && line[strlen(line)-1] != '\r' ) {
        maxlength += 512; 
        line = (char*)realloc(line, sizeof(char)*(maxlength + 1));
        fgets(addline, 512, infile);
        strcat(line, addline);
      }
      result.speciesnum = 0;
      for ( i = 0; i < strlen(line); i++ ) {
        if ( line[i] == '\t' ) {
          result.speciesnum++;
        }
      }
      free(line);
      maxlength = 7*result.speciesnum + 256;
      result.species = (char**) malloc(sizeof(char*)*result.speciesnum);
      result.distance = (float**) malloc(sizeof(float*)*result.speciesnum);
      line = (char*) malloc(sizeof(char)*(maxlength + 1));
      for ( i = 0; i < result.speciesnum; i++ ) {
        result.distance[i] = (float*) malloc(sizeof(float)*result.speciesnum);
        result.distance[i][i] = 0;
        if( !fgets(line, maxlength, infile) ) {
          perror("Wrong format\n");
          exit(1);
        }
        lenline = strlen(line);
        k = 1;
        for ( j = 0; j < result.speciesnum; j++ ) {
          p = 0;
          while ( (line[k] != '\t') && ( k < lenline) && ( p < 8 )) {
            tmpline[p] = line[k];
            p++;
            k++;
          }
          if ( (k == lenline) || (p == 8) ) {
            perror("Wrong format\n");
            exit(1);
          }
          tmpline[p] = '\0';
          k++;
          if ( i < j) {
            if ( tmpline[0] == 'n') result.distance[i][j] = 999.9999;
            else result.distance[i][j] = atof(tmpline);
          }
          else if ( j < i ) {
           result.distance[i][j] = result.distance[j][i];
          } 
        } /* for j */
        result.species[i] = (char*) malloc(sizeof(char)*(lenline - k));
        sscanf(line+k,"%s %d", result.species[i], &controlnumber);
        if ( controlnumber != i + 1 ) {
          fprintf(stderr, "contolnumber (%d) does not match i (%d)\n", controlnumber, i);
        } 
      } /* for i */
    }
  }
  return result;
}

struct tree upgma(struct distancematrix inmatrix) {
  struct tree result;
  unsigned int *active;
  float **curmat;
  float *rost;
  float min, sum;
  unsigned int i, j, k, m, n, p;
  unsigned int sp1, sp2;

  result.leavesnum = inmatrix.speciesnum;
  result.branchnum = 2*inmatrix.speciesnum - 3;
  result.leaf = (char**)malloc(sizeof(char*)*inmatrix.speciesnum);
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*result.branchnum);
  result.length = (float*)malloc(sizeof(float)*result.branchnum );
  active = (unsigned int*)malloc(sizeof(unsigned int)*result.branchnum);
  curmat = (float**)malloc(sizeof(float*)*result.branchnum);
  rost = (float*)malloc(sizeof(float)*(result.branchnum + 1));

  for ( i = 0; i < result.leavesnum; i++ ) {
    result.leaf[i] = (char*)malloc(sizeof(char)*(strlen(inmatrix.species[i]+1)));
    strcpy(result.leaf[i], inmatrix.species[i]);
  }

  for ( j = 0; j < result.branchnum; j++ ) {
    result.branch[j] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
    curmat[j] = (float*)malloc(sizeof(float)*result.branchnum);
    rost[j] = 0;
    if ( j < result.leavesnum ) {
      active[j] = 1;
      for ( i = 0; i < result.leavesnum; i++ ) {
        if ( i == j ) {
          result.branch[j][i] = 1;
        }
        else result.branch[j][i] = 0;
        curmat[j][i] = inmatrix.distance[j][i];
      }
    }
    else active[j] = 0;
  }

  n = result.leavesnum;
  if ( n == 2 ) {
    result.length[0] = inmatrix.distance[0][1];
    result.rootlocation = inmatrix.distance[0][1]/2;
    result.root = 0;
  }
  while ( n <= result.branchnum ) {
    for ( i = 0; (i < result.branchnum) && (active[i] == 0); i++);
    for ( j = i+1; (j < result.branchnum) && (active[j] == 0); j++);
    min = curmat[i][j];
    sp1 = i;
    sp2 = j;
    for ( i = sp1; i < result.branchnum; i++) { 
      if ( active[i] ) {
        for ( j = i+1; j < result.branchnum; j++) {
          if ( active[j] && (curmat[i][j] < min) ) {
            min = curmat[i][j];
            sp1 = i;
            sp2 = j; 
          }
        } /* for j */
      }
    } /* for i */
    result.length[sp1] = min/2 - rost[sp1];
    result.length[sp2] = min/2 - rost[sp2];
    rost[n] = min/2;
    active[sp1] = 0;
    active[sp2] = 0;
    if ( n < result.branchnum ) {
      active[n] = 1;
      for ( k = 0; k < result.leavesnum; k++) {
        if ( result.branch[sp1][k] == 1 || result.branch[sp2][k] == 1 ) result.branch[n][k] = 1;
        else result.branch[n][k] = 0;
      }
      for ( i = 0; i < n; i++) {
        if ( active[i] ) {
          m = 0;
          sum = 0.0;
          for ( k = 0; k < result.leavesnum; k++ ) {
            if ( result.branch[i][k] == 1 ) {
              for ( p = 0; p <  result.leavesnum; p++ ) {
                if ( result.branch[n][p] == 1 ) {
                  m++;
                  sum += inmatrix.distance[k][p];
                }
              }
            }
          } /* for k */
          curmat[i][n] = sum/m;       
        } /* if i is active */
      } /* for i */
    } 
    else {
      for ( i = 0; (i < result.branchnum) && !active[i]; i++);
      result.root = i;
      if ( i == result.branchnum ) exit(8);
      m = 0;
      sum = 0.0;
      for ( k = 0; k < result.leavesnum; k++ ) {
        if ( result.branch[i][k] == 1 ) {
          for ( p = 0; p <  result.leavesnum; p++ ) {
            if ( result.branch[sp1][p] == 1 || result.branch[sp2][p] == 1) {
              m++;
              sum += inmatrix.distance[k][p];
            }
          } 
        }
      }
      result.length[i] = sum/m - rost[n] - rost[i];
      if ( result.branch[i][0] == 1 ) {
        result.rootlocation = sum/(2*m) - rost[i];
      }
      else {
        result.rootlocation = sum/(2*m) - rost[n];
      }
    }
    n++;
  }
  result.rooted = 1;
  
  return result;
} /* upgma */

struct tree wpgma(struct distancematrix inmatrix, float **outweight) {
  struct tree result;
  unsigned int *active;
  float **curmat;
  float *rost;
  float *weight;
  unsigned int *volume;
  float min, sumweights, sumdist, ipart, npart;
  unsigned int i, j, k, n, p;
  unsigned int sp1, sp2;

  result.leavesnum = inmatrix.speciesnum;
  result.branchnum = 2*inmatrix.speciesnum - 3;
  result.leaf = (char**)malloc(sizeof(char*)*inmatrix.speciesnum);
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*result.branchnum);
  result.length = (float*)malloc(sizeof(float)*result.branchnum );
  active = (unsigned int*)malloc(sizeof(unsigned int)*result.branchnum);
  curmat = (float**)malloc(sizeof(float*)*result.branchnum);
  rost = (float*)malloc(sizeof(float)*(result.branchnum + 1));
  weight = (float*)malloc(sizeof(float)*inmatrix.speciesnum);
  volume = (unsigned int*)malloc(sizeof(unsigned int)*(result.branchnum + 1));

  for ( i = 0; i < result.leavesnum; i++ ) {
    result.leaf[i] = (char*)malloc(sizeof(char)*(strlen(inmatrix.species[i]+1)));
    strcpy(result.leaf[i], inmatrix.species[i]);
    weight[i] = 1.0;
  }
  for ( j = 0; j < result.branchnum; j++ ) {
    result.branch[j] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
    curmat[j] = (float*)malloc(sizeof(float)*result.branchnum);
    rost[j] = 0;
    volume[j] = 1;
    if ( j < result.leavesnum ) {
      active[j] = 1;       /* in beginning, active are only leaf branches */
      for ( i = 0; i < result.leavesnum; i++ ) {
        if ( i == j ) {
          result.branch[j][i] = 1;
        }
        else result.branch[j][i] = 0;
        curmat[j][i] = inmatrix.distance[j][i]; /* initial distances between trivial clusters */
      }
    }
    else active[j] = 0;
  }

  n = result.leavesnum;
  if ( n == 2 ) {
    result.length[0] = inmatrix.distance[0][1];
    result.rootlocation = inmatrix.distance[0][1]/2;
    weight[0] = weight[1] = 1.0;
    result.root = 0;
  }

  while ( n <= result.branchnum ) { 
    for ( i = 0; (i < result.branchnum) && (active[i] == 0); i++);
    for ( j = i+1; (j < result.branchnum) && (active[j] == 0); j++); /* find a pair of active branches=clusters */
    min = curmat[i][j]; /* distance between clusters */
    sp1 = i;
    sp2 = j;
    for ( i = sp1; i < result.branchnum; i++) {
      if ( active[i] ) {
        for ( j = i+1; j < result.branchnum; j++) {
          if ( active[j] && (curmat[i][j] < min) ) { /* for all pairs of active branches, finding the minimal distance  */
            min = curmat[i][j];
            sp1 = i;
            sp2 = j;
          }
        } /* for j */
      }
    } /* for i */
    result.length[sp1] = min/2 - rost[sp1];
    result.length[sp2] = min/2 - rost[sp2];
    rost[n] = min/2;
    volume[n] = volume[sp1] + volume[sp2];
    active[sp1] = 0;  /* make joined branches inactive */
    active[sp2] = 0;
    for ( k = 0; k < result.leavesnum; k++ ) { /* re-computing weights */
      if ( result.branch[sp1][k] == 1 ) {
        if ( volume[sp1] == 1 ) {
          weight[k] = result.length[sp1];
        }
        else {
          weight[k] += result.length[sp1]/volume[sp1];
        }
      }
      if ( result.branch[sp2][k] == 1 ) {
        if ( volume[sp2] == 1 ) {
          weight[k] = result.length[sp2];
        }
        else {
          weight[k] += result.length[sp2]/volume[sp2];
        }
      }
    }
    if ( n < result.branchnum ) { /* not the final cluster => just a branch */
      active[n] = 1;
      for ( k = 0; k < result.leavesnum; k++) {
        if ( result.branch[sp1][k] == 1 || result.branch[sp2][k] == 1 ) result.branch[n][k] = 1; /* join sp1 and sp2 to n */
        else result.branch[n][k] = 0;
      }
      for ( i = 0; i < n; i++) {
        if ( active[i] ) {
          sumweights = 0.0;
          sumdist = 0.0;
          for ( k = 0; k < result.leavesnum; k++ ) {
            if ( result.branch[i][k] == 1 ) {
              for ( p = 0; p <  result.leavesnum; p++ ) {
                if ( result.branch[n][p] == 1 ) {
                  sumweights += weight[k]*weight[p];
                  sumdist += inmatrix.distance[k][p]*weight[k]*weight[p];
                }
              }
            }
          } /* for k */
          curmat[i][n] = sumdist/sumweights;
        } /* if i is active */
      } /* for i */
    }
    else { /* the final cluster: rooting */
      for ( i = 0; (i < result.branchnum) && !active[i]; i++); /* finding the last active branch - it is the root */
      result.root = i;
      if ( i == result.branchnum ) exit(8); /* no active branches - something's wrong */
      sumweights = 0.0;
      sumdist = 0.0;
      for ( k = 0; k < result.leavesnum; k++ ) { /* computing tha distance between the two remaining clusters (i and n)*/
        if ( result.branch[i][k] == 1 ) {
          for ( p = 0; p <  result.leavesnum; p++ ) {
            if ( result.branch[sp1][p] == 1 || result.branch[sp2][p] == 1) {
              sumweights += weight[k]*weight[p];
              sumdist += inmatrix.distance[k][p]*weight[k]*weight[p];
            }
          }
        }
      }
      ipart = sumdist/(2*sumweights) - rost[i];
      npart = sumdist/(2*sumweights) -  rost[n];
      result.length[i] = ipart + npart;
      if ( result.branch[i][0] == 1 ) {
        result.rootlocation = ipart;
      }
      else {
        result.rootlocation = npart;
      }
      for ( k = 0; k < result.leavesnum; k++ ) {
         if ( result.branch[i][k] == 1 ) {
           weight[k] += ipart/volume[i];
         }
         if ( result.branch[sp1][p] == 1 || result.branch[sp2][p] == 1 ) {
           weight[k] += npart/volume[n];
         }
      }
    }
    n++;
  }
  result.rooted = 1;
  (*outweight) = weight;
  return result;
} /* wpgma */

struct tree wpgma2(struct distancematrix inmatrix) {
  struct tree result;
  unsigned int *active;
  float **curmat;
  float *rost;
  float *weight;
  float min, sumweights, sumdist, ipart, npart;
  unsigned int i, j, k, n, p;
  unsigned int sp1, sp2;

  result.leavesnum = inmatrix.speciesnum;
  result.branchnum = 2*inmatrix.speciesnum - 3;
  result.leaf = (char**)malloc(sizeof(char*)*inmatrix.speciesnum);
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*result.branchnum);
  result.length = (float*)malloc(sizeof(float)*result.branchnum );
  active = (unsigned int*)malloc(sizeof(unsigned int)*result.branchnum);
  curmat = (float**)malloc(sizeof(float*)*result.branchnum);
  rost = (float*)malloc(sizeof(float)*(result.branchnum + 1));
  weight = (float*)malloc(sizeof(float)*inmatrix.speciesnum);

  for ( i = 0; i < result.leavesnum; i++ ) {
    result.leaf[i] = (char*)malloc(sizeof(char)*(strlen(inmatrix.species[i]+1)));
    strcpy(result.leaf[i], inmatrix.species[i]);
    weight[i] = 1.0;
  }
  for ( j = 0; j < result.branchnum; j++ ) {
    result.branch[j] = (unsigned int*)malloc(sizeof(unsigned int)*result.leavesnum);
    curmat[j] = (float*)malloc(sizeof(float)*result.branchnum);
    rost[j] = 0;
    if ( j < result.leavesnum ) {
      active[j] = 1;       /* in beginning, active are only leaf branches */
      for ( i = 0; i < result.leavesnum; i++ ) {
        if ( i == j ) {
          result.branch[j][i] = 1;
        }
        else result.branch[j][i] = 0;
        curmat[j][i] = inmatrix.distance[j][i]; /* initial distances between trivial clusters */
      }
    }
    else active[j] = 0;
  }

  n = result.leavesnum;
  if ( n == 2 ) {
    result.length[0] = inmatrix.distance[0][1];
    result.rootlocation = inmatrix.distance[0][1]/2;
    weight[0] = weight[1] = 1.0;
    result.root = 0;
  }
  while ( n <= result.branchnum ) {
    for ( i = 0; (i < result.branchnum) && (active[i] == 0); i++);
    for ( j = i+1; (j < result.branchnum) && (active[j] == 0); j++); /* find a pair of active branches=clusters */
    min = curmat[i][j]; /* distance between clusters */
    sp1 = i;
    sp2 = j;
    for ( i = sp1; i < result.branchnum; i++) {
      if ( active[i] ) {
        for ( j = i+1; j < result.branchnum; j++) {
          if ( active[j] && (curmat[i][j] < min) ) { /* for all pairs of active branches, finding the minimal distance  */
            min = curmat[i][j];
            sp1 = i;
            sp2 = j;
          }
        } /* for j */
      }
    } /* for i */
    result.length[sp1] = min/2 - rost[sp1];
    result.length[sp2] = min/2 - rost[sp2];
    rost[n] = min/2;
    active[sp1] = 0;  /* make joined branches inactive */
    active[sp2] = 0;
    for ( k = 0; k < result.leavesnum; k++ ) { /* re-computing weights */
      if ( result.branch[sp1][k] == 1 || result.branch[sp2][k] == 1) {
        weight[k] = weight[k]/2;
      }
    }
    if ( n < result.branchnum ) { /* not the final cluster => just a branch */
      active[n] = 1;
      for ( k = 0; k < result.leavesnum; k++) {
        if ( result.branch[sp1][k] == 1 || result.branch[sp2][k] == 1 ) result.branch[n][k] = 1; /* join sp1 and sp2 to n */
        else result.branch[n][k] = 0;
      }
      for ( i = 0; i < n; i++) {
        if ( active[i] ) {
          sumweights = 0.0;
          sumdist = 0.0;
          for ( k = 0; k < result.leavesnum; k++ ) {
            if ( result.branch[i][k] == 1 ) {
              for ( p = 0; p <  result.leavesnum; p++ ) {
                if ( result.branch[n][p] == 1 ) {
                  sumweights += weight[k]*weight[p];
                  sumdist += inmatrix.distance[k][p]*weight[k]*weight[p];
                }
              }
            }
          } /* for k */
          curmat[i][n] = sumdist/sumweights;
        } /* if i is active */
      } /* for i */
    }
    else { /* the final cluster: rooting */
      for ( i = 0; (i < result.branchnum) && !active[i]; i++); /* finding the last active branch - it is the root */
      result.root = i;
      if ( i == result.branchnum ) exit(8); /* no active branches - something's wrong */
      sumweights = 0.0;
      sumdist = 0.0;
      for ( k = 0; k < result.leavesnum; k++ ) { /* computing tha distance between the two remaining clusters (i and n)*/
        if ( result.branch[i][k] == 1 ) {
          for ( p = 0; p <  result.leavesnum; p++ ) {
            if ( result.branch[sp1][p] == 1 || result.branch[sp2][p] == 1) {
              sumweights += weight[k]*weight[p];
              sumdist += inmatrix.distance[k][p]*weight[k]*weight[p];
            }
          }
        }
      }
      ipart = sumdist/(2*sumweights) - rost[i];
      npart = sumdist/(2*sumweights) -  rost[n];
      result.length[i] = ipart + npart;
      if ( result.branch[i][0] == 1 ) {
        result.rootlocation = ipart;
      }
      else {
        result.rootlocation = npart;
      }
    }
    n++;
  }
  result.rooted = 1;
  return result;
} /* wpgma2 */

unsigned int whichsplitmat(unsigned int a, unsigned int b, unsigned int c, unsigned int d, struct distancematrix matrix) {
  unsigned int result = 0;

  if ( a >= matrix.speciesnum ) {
    perror("whichsplitmat: a is out of range");
    exit(1);
  }
  if ( b >= matrix.speciesnum ) {
    perror("whichsplitmat: b is out of range");
    exit(1);
  }
  if ( c >= matrix.speciesnum ) {
    perror("whichsplitmat: c is out of range");
    exit(1);
  }
  if ( d >= matrix.speciesnum ) {
    perror("whichsplitmat: d is out of range");
    exit(1);
  }

  if ( matrix.distance[a][b] + matrix.distance[c][d] <  matrix.distance[a][c] + matrix.distance[b][d] && 
       matrix.distance[a][b] + matrix.distance[c][d] <  matrix.distance[a][d] + matrix.distance[b][c] ) {
    result = 2;
  }
  else if ( matrix.distance[a][c] + matrix.distance[b][d] <  matrix.distance[a][b] + matrix.distance[c][d] &&
       matrix.distance[a][c] + matrix.distance[b][d] <  matrix.distance[a][d] + matrix.distance[b][c] ) {
    result = 3;
  }
  else if ( matrix.distance[a][d] + matrix.distance[b][c] <  matrix.distance[a][c] + matrix.distance[b][d] &&
       matrix.distance[a][d] + matrix.distance[b][c] <  matrix.distance[a][b] + matrix.distance[c][d] ) {
    result = 4;
  }

  return result;
}

unsigned int whichsplittree(unsigned int a, unsigned int b, unsigned int c, unsigned int d, struct tree intree) {
  unsigned int result = 0;
  unsigned int i;

  if ( a >= intree.leavesnum ) {
    perror("whichsplittree: a is out of range");
    exit(1);
  }
  if ( b >= intree.leavesnum ) {
    perror("whichsplittree: b is out of range");
    exit(1);
  }
  if ( c >= intree.leavesnum ) {
    perror("whichsplittree: c is out of range");
    exit(1);
  }
  if ( d >= intree.leavesnum ) {
    perror("whichsplittree: d is out of range");
    exit(1);
  }

  for ( i = 0; i < intree.branchnum && result == 0; i++ ) {
    if ( intree.branch[i][a] == intree.branch[i][b] 
      && intree.branch[i][a] != intree.branch[i][c] 
      && intree.branch[i][c] == intree.branch[i][d] ) {
      result = 2;
    }
    else if ( intree.branch[i][a] == intree.branch[i][c] 
           && intree.branch[i][a] != intree.branch[i][b] 
           && intree.branch[i][b] == intree.branch[i][d] ) {
      result = 3;
    }
    else if ( intree.branch[i][a] == intree.branch[i][d] 
           && intree.branch[i][a] != intree.branch[i][b] 
           && intree.branch[i][b] == intree.branch[i][c] ) {
      result = 4;
    }
  }
 
  return result;
}

int topocriterium(struct distancematrix inmatrix, struct tree intree) {
  int result = 0;
  int *corresp;
  unsigned int i, j;
  unsigned int a, b, c, d;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-3; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum-2; b++ )
    for ( c = b + 1; c < inmatrix.speciesnum-1; c++ )
    for ( d = c + 1; d < inmatrix.speciesnum; d++ ) { /* for all 4-ths of species */
      if ( whichsplitmat(a, b, c, d, inmatrix) 
           == whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree) ) {
        result++;
      }
    } /* for all 4-ths */
  } /* if */

  return result;
}

/***********************************************************************
*  additcriterium: for a distance matrix and a tree on the same species 
*  returns (normalized) sum over all fourths
*  of the value DS1 + DS2 - 2*DS3
*  where DS1, DS2 - two "senior" sums of distances
*  DS3 - "junior" sum of disances (seniority according to the tree!)
*************************************************************************/
float additcriterium(struct distancematrix inmatrix, struct tree intree) {
  float result = 0.0;
  float base = 0.0;
  float resultplus;
  int *corresp;
  unsigned int i, j;
  unsigned int a, b, c, d, split;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-3; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum-2; b++ )
    for ( c = b + 1; c < inmatrix.speciesnum-1; c++ )
    for ( d = c + 1; d < inmatrix.speciesnum; d++ ) { /* for all 4-ths of species */
      base += (inmatrix.distance[a][b] + inmatrix.distance[c][d]
             + inmatrix.distance[a][c] + inmatrix.distance[a][d] 
             + inmatrix.distance[b][c] + inmatrix.distance[b][d])/6; 
      split = whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree);
      if ( split == 2 ) {
        resultplus = inmatrix.distance[a][c] + inmatrix.distance[b][d] 
                + inmatrix.distance[b][c] + inmatrix.distance[a][d] 
                - 2*(inmatrix.distance[a][b] + inmatrix.distance[c][d]);
      }
      else if ( split == 3 ) {
        resultplus = inmatrix.distance[a][b] + inmatrix.distance[a][d]
                + inmatrix.distance[b][c] + inmatrix.distance[c][d]
                - 2*(inmatrix.distance[a][c] + inmatrix.distance[b][d]);

      }
      else if ( split == 4 ) {
        resultplus = inmatrix.distance[a][b] + inmatrix.distance[a][c]
                + inmatrix.distance[b][d] + inmatrix.distance[c][d]
                - 2*(inmatrix.distance[a][d] + inmatrix.distance[b][c]);
      }
      result += resultplus;
/*
      if ( whichsplitmat(a, b, c, d, inmatrix)
           == whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree) ) {
        result++;
      }
*/
    } /* for all 4-ths */
  } /* if */
  if ( base == 0.0 ) base = 1.0;
  result = result/base;
  return result;
} /* additcriterium */

/***********************************************************************
*  additcriterium2: for a distance matrix and a tree on the same species 
*  returns (normalized) sum over all fourths
*  of the value DS1 + DS2
*  where DS1, DS2 - two "senior" sums of distances
*  (seniority according to the tree!)
*************************************************************************/
float additcriterium2(struct distancematrix inmatrix, struct tree intree) {
  float result = 0.0;
  float base = 0.0;
  float resultplus;
  int *corresp;
  unsigned int i, j;
  unsigned int a, b, c, d, split;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-3; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum-2; b++ )
    for ( c = b + 1; c < inmatrix.speciesnum-1; c++ )
    for ( d = c + 1; d < inmatrix.speciesnum; d++ ) { /* for all 4-ths of species */
      base += (inmatrix.distance[a][b] + inmatrix.distance[c][d]
             + inmatrix.distance[a][c] + inmatrix.distance[a][d] 
             + inmatrix.distance[b][c] + inmatrix.distance[b][d])/6; 
      split = whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree);
      if ( split == 2 ) {
        resultplus = (inmatrix.distance[a][c] + inmatrix.distance[b][d] 
                + inmatrix.distance[b][c] + inmatrix.distance[a][d])/4;
      }
      else if ( split == 3 ) {
        resultplus = (inmatrix.distance[a][b] + inmatrix.distance[a][d]
                + inmatrix.distance[b][c] + inmatrix.distance[c][d])/4;

      }
      else if ( split == 4 ) {
        resultplus = (inmatrix.distance[a][b] + inmatrix.distance[a][c]
                + inmatrix.distance[b][d] + inmatrix.distance[c][d])/4;
      }
      result += resultplus;
/*
      if ( whichsplitmat(a, b, c, d, inmatrix)
           == whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree) ) {
        result++;
      }
*/
    } /* for all 4-ths */
  } /* if */
  if ( base == 0.0 ) base = 1.0;
  result = result/base;
  return result;
} /* additcriterium2 */

/***********************************************************************
*  additcriterium3: for a distance matrix and a tree on the same species 
*  returns (normalized and opposed, i.e., negative) sum over all fourths
*  of the value |DS1 - DS2|
*  where DS1, DS2 - two "senior" sums of distances
*  (seniority according to the tree!)
*************************************************************************/
float additcriterium3(struct distancematrix inmatrix, struct tree intree) {
  float result = 0.0;
  float base = 0.0;
  float resultplus;
  float x, y;
  int *corresp;
  unsigned int i, j;
  unsigned int a, b, c, d, split;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-3; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum-2; b++ )
    for ( c = b + 1; c < inmatrix.speciesnum-1; c++ )
    for ( d = c + 1; d < inmatrix.speciesnum; d++ ) { /* for all 4-ths of species */
      base += (inmatrix.distance[a][b] + inmatrix.distance[c][d]
             + inmatrix.distance[a][c] + inmatrix.distance[a][d] 
             + inmatrix.distance[b][c] + inmatrix.distance[b][d])/6; 
      split = whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree);
      if ( split == 2 ) {
        x = inmatrix.distance[a][c] + inmatrix.distance[b][d];
        y = inmatrix.distance[b][c] + inmatrix.distance[a][d];
      }
      else if ( split == 3 ) {
        x = inmatrix.distance[a][b] + inmatrix.distance[a][d];
        y = inmatrix.distance[b][c] + inmatrix.distance[c][d];

      }
      else if ( split == 4 ) {
        x = inmatrix.distance[a][b] + inmatrix.distance[a][c];
        y = inmatrix.distance[b][d] + inmatrix.distance[c][d];
      }
      if ( x > y ) {
        if ( y > 0 ) result += x - y;
      }
      else {
        if ( x > 0 ) result += y - x;
      }
    } /* for all 4-ths */
  } /* if */
  if ( base == 0.0 ) base = 1.0;
  result = -result/base;
  return result;
} /* additcriterium3 */

/***********************************************************************
*  additcriterium4: for a distance matrix and a tree on the same species 
*  returns (normalized and opposed, i.e., negative) sum over all fourths
*  of the value |DS1 - DS2|/min(DS1, DS2)
*  where DS1, DS2 - two "senior" sums of distances
*  (seniority according to the tree!)
*************************************************************************/
float additcriterium4(struct distancematrix inmatrix, struct tree intree) {
  float result = 0.0;
  float resultplus;
  float x, y;
  int *corresp;
  unsigned int i, j, n;
  unsigned int a, b, c, d, split;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-3; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum-2; b++ )
    for ( c = b + 1; c < inmatrix.speciesnum-1; c++ )
    for ( d = c + 1; d < inmatrix.speciesnum; d++ ) { /* for all 4-ths of species */
      split = whichsplittree(corresp[a], corresp[b], corresp[c],  corresp[d], intree);
      if ( split == 2 ) {
        x = inmatrix.distance[a][c] + inmatrix.distance[b][d];
        y = inmatrix.distance[b][c] + inmatrix.distance[a][d];
      }
      else if ( split == 3 ) {
        x = inmatrix.distance[a][b] + inmatrix.distance[a][d];
        y = inmatrix.distance[b][c] + inmatrix.distance[c][d];

      }
      else if ( split == 4 ) {
        x = inmatrix.distance[a][b] + inmatrix.distance[a][c];
        y = inmatrix.distance[b][d] + inmatrix.distance[c][d];
      }
      if ( x > y ) {
        if ( y > 0 ) result += (x - y)/y;
      }
      else {
        if ( x > 0 ) result += (y - x)/x;
      }
    } /* for all 4-ths */
  } /* if */
  n = inmatrix.speciesnum;
  result = -24*result/(n*(n-1)*(n-2)*(n-3));
  return result;
} /* additcriterium4 */


float bmecriterium(struct distancematrix inmatrix, struct tree intree) {
  float evol = 0.0;
  float base= 0.0;
  float result;
  unsigned int a, b, i, j;
  int *corresp;

  if ( inmatrix.speciesnum == intree.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*inmatrix.speciesnum);
    for ( i = 0; i < inmatrix.speciesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < intree.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(intree.leaf[j], inmatrix.species[i]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) return -1;
    }
    for ( a = 0; a < inmatrix.speciesnum-1; a++ )
    for ( b = a + 1; b < inmatrix.speciesnum; b++ ){
      evol += inmatrix.distance[a][b]/pow(2.0, combdistance(intree, corresp[a], corresp[b]));
      base += inmatrix.distance[a][b];
    }
  }
  if ( base == 0.0 ) result = 0.0;
  else result = - evol/base;
  return result;
} /* bmecriterium */

/******************************************************************************************** 
*  topotryaddleaf: returnes the increase of topological value ( = number of correct fourths)
*  after adding a new leaf with the name _newleaf_ (presented as a species in the _matrix_)
*  to the brunch number _branchindex_ of the tree _intree_
********************************************************************************************/
int topotryaddleaf(struct distancematrix matrix, struct tree intree, char *newleaf, unsigned int branchindex) {
  int result = 0;
  unsigned int i, j, flag, split, newleafindex;
  unsigned int a, b, c;
  unsigned int *corresp;

  corresp = (unsigned int *) calloc(intree.leavesnum, sizeof(unsigned int));
  for ( i = 0; i < intree.leavesnum; i++) {
    flag = 0;
    for ( j = 0; j < matrix.speciesnum && flag == 0; j++ ) {
      if ( strcmp ( intree.leaf[i], matrix.species[j] ) == 0 ) {
        corresp[i] = j;
        flag = 1;
      }
    } /* for j */
    if ( flag == 0 ) {
      fprintf(stderr, "topotryaddleaf: the tree contains leaf \"%s\" not presented in the matrix\n", intree.leaf[i]);
      exit(1);
    }
  } /* for i */
 
  flag = 0;
  for ( j = 0; j < matrix.speciesnum && flag == 0; j++ ) {
    if ( strcmp ( newleaf, matrix.species[j] ) == 0 ) {
      newleafindex = j;
      flag = 1;
    }
  }
  if ( flag == 0 ) {
    fprintf(stderr, "topotryaddleaf: the new leaf \"%s\" is not presented in the matrix\n", newleaf);
      exit(1);
  }

  for ( a = 0; a < intree.leavesnum-2; a++ )
  for ( b = a + 1; b < intree.leavesnum-1; b++ )
  for ( c = b + 1; c < intree.leavesnum; c++ ) {
    split = 0;
    for ( i = 0; i < intree.branchnum && split == 0; i++ ){
      if ( i != branchindex ) {
        if ( intree.branch[i][b] == intree.branch[i][c] &&  intree.branch[i][a] != intree.branch[i][b] ) {
          if ( branchbetween(intree, i, b, branchindex ) ) /* new leaf is together with a */ 
            split = 2;
        }
        else if ( intree.branch[i][a] == intree.branch[i][c] &&  intree.branch[i][b] != intree.branch[i][a] ) {
          if ( branchbetween(intree, i, a, branchindex ) ) /* new leaf is together with b */
            split = 3;
        }
        else if ( intree.branch[i][a] == intree.branch[i][b] &&  intree.branch[i][c] != intree.branch[i][a] ) {
          if ( branchbetween(intree, i, a, branchindex ) ) /* new leaf is together with c */
            split = 4;
        }
      }
      else { /* i == branchindex */
        if ( intree.branch[i][a] != intree.branch[i][b] && intree.branch[i][b] == intree.branch[i][c] )
          split = 2;
        else if ( intree.branch[i][b] != intree.branch[i][c] && intree.branch[i][c] == intree.branch[i][a] )
          split = 3;
        else if ( intree.branch[i][c] != intree.branch[i][a] && intree.branch[i][a] == intree.branch[i][b] )
          split = 4;
      }
    } /* for all branches */
    if ( split == whichsplitmat(newleafindex, corresp[a], corresp[b], corresp[c], matrix) ) {
        result++;
    }
  } /* for all triples (a, b, c) of old leaves */

  return result;
} /* topotryaddleaf */


int topotryaddleaf_b(float **distances, struct tree intree, unsigned int newleafindex, unsigned int branchindex) {
  int result = 0;
  unsigned int i, j, flag, split, split1;
  unsigned int a, b, c;
  float s2, s3, s4;

  for ( a = 0; a < intree.leavesnum-2; a++ )
  for ( b = a + 1; b < intree.leavesnum-1; b++ )
  for ( c = b + 1; c < intree.leavesnum; c++ ) {
    split = 0;
    for ( i = 0; i < intree.branchnum && split == 0; i++ ){
      if ( i != branchindex ) {
        if ( intree.branch[i][b] == intree.branch[i][c] &&  intree.branch[i][a] != intree.branch[i][b] ) {
          if ( branchbetween(intree, i, b, branchindex ) ) /* new leaf is together with a */
            split = 2;
        }
        else if ( intree.branch[i][a] == intree.branch[i][c] &&  intree.branch[i][b] != intree.branch[i][a] ) {
          if ( branchbetween(intree, i, a, branchindex ) ) /* new leaf is together with b */
            split = 3;
        }
        else if ( intree.branch[i][a] == intree.branch[i][b] &&  intree.branch[i][c] != intree.branch[i][a] ) {
          if ( branchbetween(intree, i, a, branchindex ) ) /* new leaf is together with c */
            split = 4;
        }
      }
      else { /* i == branchindex */
        if ( intree.branch[i][a] != intree.branch[i][b] && intree.branch[i][b] == intree.branch[i][c] )
          split = 2;
        else if ( intree.branch[i][b] != intree.branch[i][c] && intree.branch[i][c] == intree.branch[i][a] )
          split = 3;
        else if ( intree.branch[i][c] != intree.branch[i][a] && intree.branch[i][a] == intree.branch[i][b] )
          split = 4;
      }
    } /* for all branches */
    if ( split != 0 ) {
      split1 = 0;
      s2 = distances[newleafindex][a] + distances[b][c];
      s3 = distances[newleafindex][b] + distances[a][c];
      s4 = distances[newleafindex][c] + distances[a][b];
      if ( s2 < s3 && s2 < s4 ) split1 = 2;
      else if ( s3 < s2 && s3 < s4 ) split1 = 3;
      else if ( s4 < s2 && s4 < s3 ) split1 = 4;
    }
    if ( split == split1 ) result++;
  } /* for all triples (a, b, c) of old leaves */

  return result;
} /* topotryaddleaf_b */

struct distancematrix partmatrix(struct distancematrix inmatrix, unsigned int part) {
  struct distancematrix result;
  unsigned int i, j;

  if ( part > inmatrix.speciesnum ) {
    fprintf(stderr, "part (%d) is out of range (0 to %d)\n", part, inmatrix.speciesnum);
    exit(1);
  }
  else {
    result.speciesnum = part;
    result.species = (char**) malloc(sizeof(char*)*part);
    result.distance = (float**) malloc(sizeof(float*)*part);
    for ( i = 0; i < part; i++ ){
      result.species[i] = (char*) malloc(sizeof(char)*(strlen(inmatrix.species[i])+1));
      strcpy(result.species[i], inmatrix.species[i]);
      result.distance[i] = (float*) malloc(sizeof(float)*part);
      for ( j = 0; j < part; j++ ){
        result.distance[i][j] = inmatrix.distance[i][j];
      }
    }
  }

  return result;
}

void removetree(struct tree *totree) {
  unsigned int i, j;

  if (totree){
    if ( totree -> branch ) {  
      for ( i = 0; i < totree -> branchnum; i++ ) {
        if ( (totree -> branch)[i]) {
          free((totree -> branch)[i]);
          (totree -> branch)[i] = NULL; 
        }
      }
      free(totree -> branch);
      (totree -> branch) = NULL;
    }
    if ( totree -> length ) {
      free(totree -> length);
      (totree -> length) = NULL;
    }

    if ( totree -> leaf ) {
      for ( j = 0; j < totree -> leavesnum; j++ ) {
        if ( (totree -> leaf)[j] ) {
          free((totree -> leaf)[j]);
          (totree -> leaf)[j] = NULL;
        }
      }
      free(totree -> leaf);
      (totree -> leaf) = NULL;
    }
    totree -> branchnum = 0;
    totree -> leavesnum = 0;
    totree -> rooted = 0;
  }

  return;
} /* removetree */

void removematrix ( struct distancematrix *tomatrix ) {
  unsigned int i;

  if (tomatrix) {    
    if ( tomatrix -> distance ) {
      for ( i = 0; i < tomatrix -> speciesnum; i++ ){
        if ( (tomatrix -> distance)[i] ) {
          free( (tomatrix -> distance)[i] );
          (tomatrix -> distance)[i] = NULL;
        }
      }
      free(tomatrix -> distance);
      tomatrix -> distance = NULL;
    }

    if ( tomatrix -> species ) {
      for ( i = 0; i < tomatrix -> speciesnum; i++ ){
        if ( (tomatrix -> species)[i] ) {
          free( (tomatrix -> species)[i] );
          (tomatrix -> species)[i] = NULL;
        }
      }
      free(tomatrix -> species);
      tomatrix -> species = NULL;
    }

    tomatrix -> speciesnum = 0;
  }

  return;
} /* removematrix */

int branchdist (struct tree tree1, struct tree tree2) {
  int result = -1;
  int common;
  int *corresp;
  unsigned int i, j, k, flag, iflag;

  if ( tree1.leavesnum == tree2.leavesnum ) {
    corresp = (int*)malloc(sizeof(int)*tree1.leavesnum);
    for ( i = 0; i < tree1.leavesnum; i++ ) {
      corresp[i] = -1;
      for ( j = 0; j < tree1.leavesnum && corresp[i] == -1; j++ ) {
        if ( strcmp(tree1.leaf[i], tree2.leaf[j]) == 0 ) {
          corresp[i] = j;
        }
      }
      if ( corresp[i] == -1 ) {
        fprintf(stderr, "The leaf %s has no correspondence in the tree 2\n", tree1.leaf[i]);
        return -1;
      }
    }
 
    common = 0;
    for ( i = 0; i < tree1.branchnum; i++ ) {
      iflag = 0;
      for ( j = 0; j < tree2.branchnum && iflag == 0; j++ ) {
        flag = 0;
        for ( k = 0; k < tree1.leavesnum && flag < 3; k++ ) {
          if ( flag == 0 ) {
            if (tree1.branch[i][k] == tree2.branch[j][corresp[k]] ) {
              flag = 1;
            }
            else {
              flag = 2;
            }
          }
          else if ( flag == 1 && tree1.branch[i][k] != tree2.branch[j][corresp[k]] ) {
            flag = 3;
          }
          else if ( flag == 2 && tree1.branch[i][k] == tree2.branch[j][corresp[k]] ) {
            flag = 3;
          }
        }
        if ( flag < 3 ) { /* i and j are equal branches */
          common++;
          iflag = 1;
        }
      }
    }
    result = tree1.branchnum + tree2.branchnum - 2*common; 
  } /* if */

  return result;
} /* branchdist */


/**********************************************************
* puttree_i: pushing the tree "newtree" into the ordered 
* list "curtrees" of at max "keep" number of trees.
* Here "tokeeped" is the pointer to the number of really 
* stored trees; "qualities" is the list of tree qualities 
**********************************************************/
int puttree_i(struct tree newtree, int newquality, struct tree *curtrees, 
             unsigned int *tokeeped, unsigned int keep, int *qualities) {
  unsigned int i, j, flag;
  int result = 0;

  if ( keep == 0 ) {
    perror("Error in putree: the parameter \"keep\" should be positive\n");
    exit(1);
  }
  if ( (*tokeeped) > keep ) {
    fprintf(stderr, "Error: keeped = %d, keep = %d\n", *tokeeped, keep);
    exit(4);
  }

  if ( (*tokeeped) == 0 ) {
    result = 1;
    curtrees[0] = treecopy(newtree);
    qualities[0] = newquality;
    (*tokeeped) = 1;
  }
  else {
    i = 0;
    while ( i < (*tokeeped) && newquality < qualities[i] ) i++;
    if ( i == (*tokeeped) && (*tokeeped) < keep ) {/* just put newtree to the bottom of the list */
      result = 1;
      (*tokeeped)++;
      qualities[(*tokeeped)-1] = newquality;
      curtrees[(*tokeeped)-1] = treecopy(newtree);
    }
    else if ( i < (*tokeeped)) {
      flag = 0;
      while ( newquality == qualities[i] ) {
        if ( branchdist(newtree, curtrees[i]) == 0 ) flag = 1; 
        i++;
      }
      if ( i < keep && flag == 0) {
     /* we found new good tree */
        result = 1;
        if ( (*tokeeped) < keep ) (*tokeeped)++;
        else removetree(&(curtrees[keep-1]));
        for ( j = (*tokeeped) - 1; j > i; j-- ) {
          curtrees[j] = treecopy(curtrees[j-1]);
          removetree(&(curtrees[j-1]));
          qualities[j] = qualities[j-1];
        }
        curtrees[i] = treecopy(newtree);
        qualities[i] = newquality;
      }
    }
  }

  return result;
} /* puttree_i */

/**********************************************************
* puttree_f: pushing the tree "newtree" into the ordered 
* list "curtrees" of at max "keep" number of trees.
* Here "tokeeped" is the pointer to the number of really 
* stored trees; "qualities" is the list of tree qualities 
**********************************************************/
int puttree_f(struct tree newtree, float newquality, struct tree *curtrees, 
             unsigned int *tokeeped, unsigned int keep, float *qualities) {
  unsigned int i, j, flag;
  int result = 0;

  if ( keep == 0 ) {
    perror("Error in putree: the parameter \"keep\" should be positive\n");
    exit(1);
  }
  if ( (*tokeeped) > keep ) {
    fprintf(stderr, "Error: keeped = %d, keep = %d\n", *tokeeped, keep);
    exit(4);
  }

  if ( (*tokeeped) == 0 ) {
    result = 1;
    curtrees[0] = treecopy(newtree);
    qualities[0] = newquality;
    (*tokeeped) = 1;
  }
  else {
    i = 0;
    while ( i < (*tokeeped) && newquality < qualities[i] ) i++;
    if ( i == (*tokeeped) && (*tokeeped) < keep ) {/* just put newtree to the bottom of the list */
      result = 1;
      (*tokeeped)++;
      qualities[(*tokeeped)-1] = newquality;
      curtrees[(*tokeeped)-1] = treecopy(newtree);
    }
    else if ( i < (*tokeeped)) {
      flag = 0;
      while ( newquality == qualities[i] ) {
        if ( branchdist(newtree, curtrees[i]) == 0 ) flag = 1; 
        i++;
      }
      if ( i < keep && flag == 0) {
     /* we found new good tree */
        result = 1;
        if ( (*tokeeped) < keep ) (*tokeeped)++;
        else removetree(&(curtrees[keep-1]));
        for ( j = (*tokeeped) - 1; j > i; j-- ) {
          curtrees[j] = treecopy(curtrees[j-1]);
          removetree(&(curtrees[j-1]));
          qualities[j] = qualities[j-1];
        }
        curtrees[i] = treecopy(newtree);
        qualities[i] = newquality;
      }
    }
  }

  return result;
} /* puttree_f */

/*************************************************************
* subtree
*  for a given tree and a list of names creates the subtree
*  on species contained in the list
**************************************************************/
struct tree subtree(struct tree intree, char ** list, unsigned int listlen) {
  unsigned int i, j, k, flag;
  unsigned int *newbranches;
  unsigned int *newleaves;
  unsigned int newbranchnum, countleaf;
  int *correspleaf;
  int *correspbranch;
  struct tree result;

  result.rooted = intree.rooted;
  result.leavesnum = 0;
  result.leaf = (char**)malloc(sizeof(char*)*listlen);
  correspleaf = (int*)malloc(sizeof(int)*listlen);
  correspbranch = (int*)malloc(sizeof(int)*intree.branchnum);
  newleaves = (unsigned int*)calloc(intree.leavesnum, sizeof(unsigned int));
  i = j = 0;
  while ( i < listlen) {
    flag = 0;
    for ( k = 0; k < intree.leavesnum && flag == 0; k++) {
      if ( strcmp(intree.leaf[k], list[i]) == 0 ) {
        flag = 1;
        newleaves[k] = 1;
        result.leavesnum++;
        result.leaf[j] = (char*)malloc(sizeof(char)*(strlen(list[i])+1));
        strcpy(result.leaf[j], list[i]);
        correspleaf[j] = k;
        j++;
      } /* if */
    } /* for k */
    i++;
  } /* while i<listlen */

  newbranches = (unsigned int*)calloc(intree.branchnum, sizeof(unsigned int));
  newbranchnum = 0;
  for ( i = 0; i < intree.branchnum; i++ ) {
    newbranches[i] = 1; 
    countleaf = 0;
    for ( k = 0; k < intree.leavesnum; k++ ) {
      if (newleaves[k] == 1 ) countleaf += intree.branch[i][k];
    }
    if ( countleaf > 0 && countleaf < result.leavesnum ) {
      flag = 3;
      for ( j = 0; j < i && flag == 3; j++ ) {
        if ( newbranches[j] == 1 ) {
          flag = 0;
          
          for ( k = 0; k < intree.leavesnum && flag < 3; k++ ) {
            if (newleaves[k] == 1 ) {
              if ( flag == 0 && intree.branch[i][k] == intree.branch[j][k] ) flag = 1;
              else if ( flag == 0 && intree.branch[i][k] != intree.branch[j][k] ) flag = 2;
              else if ( flag == 1 && intree.branch[i][k] != intree.branch[j][k] ) flag = 3;
              else if ( flag == 2 && intree.branch[i][k] == intree.branch[j][k] ) flag = 3;
            } /* if */
          } /* for k */
          if ( flag < 3 ) {
            newbranches[i] = 0;
            correspbranch[i] = correspbranch[j];
          }
        } /* if newbranches[j] == 1 */
      } /* for j */
    } /* if */
    else { 
      newbranches[i] = 0;
      correspbranch[i] = -1; 
    }
    if ( newbranches[i] == 1 ) {
      correspbranch[i] = newbranchnum;
      newbranchnum++;
    }
  }
  result.branchnum = newbranchnum;
  printf("newbranchnum = %d\n", newbranchnum);
  result.branch = (unsigned int**)malloc(sizeof(unsigned int*)*newbranchnum);

  i = j = 0;
  while ( i < newbranchnum ) {
    if ( newbranches[j] == 1 ) {
      result.branch[i] = (unsigned int*)calloc(result.leavesnum, sizeof(int));
      for ( k = 0; k < result.leavesnum; k++ ) {
        result.branch[i][k] = intree.branch[j][correspleaf[k]];
      }

      i++;
    }
    j++;
  }

  if ( intree.rooted ) {
    if ( correspbranch[intree.root] >= 0 ) {
      result.root = correspbranch[intree.root];
    }
    else {
      result.root = 0; /* !!! Subject to change !!! */
    }
  }

  result.phylogram = intree.phylogram;
  result.length = (float*)calloc(result.branchnum, sizeof(float));
  if ( intree.phylogram ) {
    for ( j = 0; j < intree.branchnum; j++ ) {
      if ( correspbranch[j] >= 0 ) result.length[correspbranch[j]] += intree.length[j];
    }
    if ( intree.rooted ) {
      result.rootlocation = intree.rootlocation; /* !!! Subject to change !!! */
    }
  }
  else {
    for ( i = 0; i < result.branchnum; i++ ) {
      result.length[i] = 1.0;
    }
    result.rootlocation = 1.0;
  }

  return result;
} /* subtree */

char *writephylipmatrix(struct distancematrix inmatrix) {
  char *result, *distance;
  int size;
  int i, j, k, l;

  size = 6 + 10*inmatrix.speciesnum + 10*inmatrix.speciesnum*inmatrix.speciesnum + (inmatrix.speciesnum%7+1)*inmatrix.speciesnum; 
  result = (char *) malloc ( sizeof(char) * size);
  distance = (char *) malloc ( sizeof(char) * 10);

  sprintf(result,"%5d\n", inmatrix.speciesnum);

  for( i = 0; i < inmatrix.speciesnum; i++ ) {
    strcat(result, inmatrix.species[i]);
    for ( l = 0; l < 10 - strlen(inmatrix.species[i]); l++ ) {
      strcat(result, " ");
    }
    k = 1;
    for ( j = 0; j < inmatrix.speciesnum; j++ ) {
      sprintf(distance, "%10.6f", inmatrix.distance[i][j]);
      strcat(result, distance);
      k++;
      if ( k%7 == 0 && inmatrix.speciesnum - j > 1 ) {
        strcat(result, "\n");
        k = 0;
      } /* if */
    } /* for j */
    strcat(result, "\n");
  } /* for i */

  return result;
}

